/*
 ** FlashManager
 *
 * Copyright 2004 Donald Kinzer, All rights reserved.
 *
 * Permission is hereby granted to all persons to use this software
 * for private purposes and educational purposes.  All commercial use
 * is prohibited except pursuant to a duly executed license agreement.
 * Parties interested in licensing this software may contact the author
 * using the contact information below.
 *
 *  Donald Kinzer
 *  503-524-9355
 *  don@kinzers.com
 *  dkinzer@usa.com
 *
 * This program is provided as is and without any warranty whatsoever.
 * Anyone using the program assumes full responsibility for judging its
 * efficacy, fitness and safety for whatever use they choose.
 *
 *
 * Purpose:
 *
 * This program was written to interface with a Flash memory reader/
 * writer.
 *
 * Invocation syntax:
 *
 * FlashManager	<option>* <command>+
 *
 * where <option> is one of
 *
 * -c<comspec>    specifies the COM port, speed, etc., e.g.
 *                "COM1:,19200,N,8,1" specifies COM1 at 19200 baud,
 *                no parity, 8 data bits and 1 stop bit
 *
 * where <command> is one of
 *
 * -e             erase the Flash device
 *
 * -r[<filename>] read the Flash device and echo Intel hex records to
 *                stdout or the specified file
 *
 * -w<filename>   write the Intel hex records in the specified file to
 *                the Flash device
 *
 * -v<filename>   compare the content of the Flash device to the Intel
 *                hex records in the specified file
 *
 * -h             displays invocation syntax
 *
 *
 */

/** include files **/
#ifdef WIN32
#include <windows.h>
#endif
#include <stdio.h>
#include <io.h>
#include <ctype.h>
#include <limits.h>

/** local definitions **/

#define DEF_COMM				"COM1:38400,N,8,1"

#define HEX_READ_PROMPT			0
#define HEX_READ_START			1
#define HEX_READ_BYTECNT		2
#define HEX_READ_ADDRESS		3
#define HEX_READ_RECTYPE		4
#define HEX_READ_RECORD			5
#define HEX_READ_CHKSUM			6
#define HEX_READ_DONE			100

#define HEX_WRITE_PROMPT		0
#define HEX_WRITE_RECORDS		1
#define HEX_WRITE_DONE			100

#define REC_TYPE_DATA			0
#define REC_TYPE_END			1
#define REC_TYPE_INVALID		0xff

#define MAX_REC_LEN				43		// 16 bytes plus overhead

// programmer communication protocol characters
#define NO_ACK					'\0'
#define ESC_CHAR				0x1b
#define ACK_CHAR				'A'
#define NAK_CHAR				'N'
#define MISMATCH_CHAR			'M'

#define ERASE_CMD				'E'
#define READ_CMD				'R'
#define WRITE_CMD				'W'
#define VERIFY_CMD				'V'

/** private data **/
static LPCSTR ver = "v0.1";
static LPCSTR commSpec = DEF_COMM;

static LPCSTR helpStrings[] =
{
	"Invocation syntax:",
	"",
	"FlashManager <option>* <command>+",
	"",
	"where <option> is one of:",
	"",
	"-c<comspec>    specifies the COM port, speed, etc., e.g.",
	"               COM1:,19200,N,8,1 specifies COM1 at 19200 baud,",
	"               no parity, 8 data bits and 1 stop bit",
	"",
	"where <command> is one of:",
	"",
	"-e             erase the Flash device",
	"",
	"-r[<filename>] read the Flash device and echo Intel hex records to",
	"               stdout or the specified file",
	"",
	"-w<filename>   write the Intel hex records in the specified file to",
	"               the Flash device",
	"",
	"-v<filename>   compare the content of the Flash device to the Intel",
	"               hex records in the specified file",
	"",
	"-h             displays invocation syntax",
	NULL
};

/** internal functions **/
static void abortCommand(HANDLE cHand);
static HANDLE commInit(LPCSTR commSpec, int testOnly);
static HANDLE commDeInit(HANDLE hand);
static void displayHelp(void);
static int eraseFlash(void);
static BYTE hexVal(BYTE c);
static int readFlash(FILE *fp);
static BYTE validateHexRecord(BYTE *buf, int *lenp);
static int writeFlash(FILE *fp, int verifyOnly);

/** callback functions **/

/** public functions **/

int
main(int argc, char **argv)
{
	char *argp;
	LPCSTR lastFile = "";

	fprintf(stderr, "FlashManager %s Copyright 2004 Donald Kinzer\n", ver);

	if (--argc <= 0)
	{
		displayHelp();
		return(0);
	}		
	argv++;
	while (argc-- > 0)
	{
		FILE *fp;
		LPCSTR p;

		argp = *argv++;

		if (*argp == '-')
		{
			char c;

			c = argp[1];
			if (isalpha(c) && isupper(c))
				c = tolower(c);
			p = argp + 2;
			switch(argp[1])
			{
			case 'c':
				if (commInit(p, TRUE) == INVALID_HANDLE_VALUE)
					exit(1);
				commSpec = p;
				break;

			case 'e':
				eraseFlash();
				break;
			
			case 'r':
				if (*p == '\0')
				{
					fp = NULL;
					lastFile = NULL;
				}
				else if ((fp = fopen(p, "w")) == 0)
				{
					fprintf(stderr, "Error opening file \"%s\" for writing flash content.\n", p);
					exit(1);
				}
				else
					lastFile = p;
				readFlash(fp);
				if (fp != NULL)
					fclose(fp);
				break;

			case 'v':
				if ((*p == '\0') && (((p = lastFile) == NULL) || (*p == '\0')))
				{
					fprintf(stderr, "Missing filename of hex file to verify.\n");
					exit(1);
				}
				if ((fp = fopen(p, "r")) == 0)
				{
					fprintf(stderr, "Error opening file \"%s\" for reading.\n", p);
					exit(1);
				}
				writeFlash(fp, TRUE);
				if (fp != NULL)
					fclose(fp);
				break;

			case 'w':
				if (*p == '\0')
				{
					fprintf(stderr, "Missing filename of hex file to write to flash.\n");
					exit(1);
				}
				if ((fp = fopen(p, "r")) == 0)
				{
					fprintf(stderr, "Error opening file \"%s\" for reading.\n", p);
					exit(1);
				}
				lastFile = p;
				writeFlash(fp, FALSE);
				if (fp != NULL)
					fclose(fp);
				break;

			case 'h':
				displayHelp();
				break;

			default:
				fprintf(stderr, "Unrecognized option - \"%s\".\n", argp);
				exit(1);
				break;
			}
		}
		else
		{
			fprintf(stderr, "Unrecognized option - \"%s\".\n", argp);
			exit(1);
			break;
		}
	}
	exit(0);
	return(0);
}


/** private functions **/

/*
 ** abortCommand
 *
 * Send an abort command to the programmer.  This is typically used
 * at the start of a command to ensure that the programmer is in command mode.
 *
 */
static void
abortCommand(HANDLE cHand)
{
	if (cHand != INVALID_HANDLE_VALUE)
	{
		char c;
		DWORD cnt;
		COMSTAT cs;

		c = ESC_CHAR;
		WriteFile(cHand, &c, 1, &cnt, NULL);

		// clean out the receive buffer
		ClearCommError(cHand, &cnt, &cs);
		while ((cs.cbInQue != 0) && ReadFile(cHand, &c, 1, &cnt, NULL))
			;
	}	
}

/*
 ** commInit
 *
 * Attempt to open a comm channel if one is not already open.  Return zero
 * if successful, non-zero otherwise.
 *
 */
static HANDLE
commInit(LPCSTR commSpec, int testOnly)
{
	HANDLE handle = INVALID_HANDLE_VALUE;
	DCB dcb;
	LPCSTR p;
	char chan[10], msgBuf[200];
	LPCSTR msg = NULL;
	DWORD err = 0;

	if (commSpec == NULL)
		return(INVALID_HANDLE_VALUE);

	// skip over the COMMn: portion of the string
	if ((p = strchr(commSpec, ':')) != NULL)
		p++;
	else
		p = commSpec;

	// get the comm parameters
	memset(&dcb, 0, sizeof(dcb));
	dcb.DCBlength = sizeof(dcb);
	if (BuildCommDCB(p, &dcb))
	{
		int ival = 0;

		// determine the comm port
		for (p = commSpec; isspace(*p); p++)
			;
		if (memicmp(p, "COM", 3) == 0)
			for (p += 3; isdigit(*p); p++)
				ival = ival * 10 + (*p - '0');
		if (ival <= 0)
		{
			wsprintf(msgBuf, "Invalid comm port string:\"%s\".\n", commSpec);
			msg = msgBuf;
			goto done;
		}

		// Form new style comm channel string - "\\.\COMn" where n is
		// the desired port number.  This allows for ports 10 and above.
		wsprintf(chan, "\\\\.\\COM%u", ival);

		// attempt to open the channel and set the state
		handle = CreateFile(chan, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
		if (handle == INVALID_HANDLE_VALUE)
		{
			wsprintf(msgBuf, "Can't open COM%u", ival);
			msg = msgBuf;
			err = GetLastError();
		}
		else if (testOnly)
			CloseHandle(handle);
		else if (SetCommState(handle, &dcb))
			// assert DTR to allow communication to commence
			EscapeCommFunction(handle, SETDTR);
		else
		{
			// some type of error occurred
			err = GetLastError();
			CloseHandle(handle);
			handle = INVALID_HANDLE_VALUE;
		}
	}
	else
	{
		wsprintf(msgBuf, "Invalid comm port string:\"%s\".", commSpec);
		msg = msgBuf;
		err = GetLastError();
	}
	if ((handle == INVALID_HANDLE_VALUE) && (msg == NULL))
	{
		wsprintf(msgBuf, "Comm port initialization failed for \"%s\", err=%u.", commSpec, err);
		msg = msgBuf;
	}
done:
	if (msg != NULL)
		fprintf(stderr, msg);
	return(handle);
}

/*
 ** commDeInit
 *
 * Clean up after a comm session.
 *
 */
static HANDLE
commDeInit(HANDLE hand)
{
	if (hand != INVALID_HANDLE_VALUE)
	{
		EscapeCommFunction(hand, CLRDTR);
		EscapeCommFunction(hand, CLRRTS);
		CloseHandle(hand);
	}
	return(INVALID_HANDLE_VALUE);
}

/*
 ** displayHelp
 *
 * Display invocation syntax.
 *
 */
static void
displayHelp(void)
{
	LPCSTR *hsp;

	for (hsp = helpStrings; *hsp != NULL; hsp++)
	{
		fputs(*hsp, stderr);
		fputs("\n", stderr);
	}
}

/*
 ** eraseFlash
 *
 * Send the programming device an erase command.
 *
 */
static int
eraseFlash(void)
{
	char c;
	int stat = -1;
	HANDLE cHand;
	DWORD cnt;

	if ((cHand = commInit(commSpec, FALSE)) == INVALID_HANDLE_VALUE)
	{
		fprintf(stderr, "Can't open comm channel using \"%s\".\n", commSpec);
		return(-1);
	}

	// make sure the programmer is in command mode
	abortCommand(cHand);

	// command the device to erase the flash memory
	c = ERASE_CMD;
	if (!WriteFile(cHand, &c, 1, &cnt, NULL))
	{
		fprintf(stderr, "Error writing to comm port.\n");
	}
	else
	{
		// wait for the ack before proceeding
		while (1)
		{
			COMSTAT cs;

			ClearCommError(cHand, &cnt, &cs);
			if ((cs.cbInQue != 0) && ReadFile(cHand, &c, 1, &cnt, NULL) && (cnt > 0))
			{
				if (c == ACK_CHAR)
				{
					fprintf(stderr, "Flash memory erased.\n");
					stat = 0;
					break;
				}
				if (c == NAK_CHAR)
				{
					fprintf(stderr, "Device aborted transmission.\n");
					break;
				}
			}
			Yield();
		}
	}
	commDeInit(cHand);
	return(stat);
}

/*
 ** hexVal
 *
 * Return the digit value of a hexadecimal character.
 *
 */
static BYTE
hexVal(BYTE c)
{
	if ((c >= '0') && (c <= '9'))
		return(c - '0');
	if ((c >= 'A') && (c <= 'F'))
		return(c - 'A' + 10);
	if ((c >= 'a') && (c <= 'f'))
		return(c - 'a' + 10);
	return(0);
}

/*
 ** readFlash
 *
 * Retrieve flash memory content from the device and write
 * it in Intel hex format to the given file handle.
 *
 */
static int
readFlash(FILE *fp)
{
	int stat = -1;
	HANDLE cHand;
	int state;
	int feedback;
	WORD recCnt = 0;

	if (fp == NULL)
		fp = stdout;

	if ((cHand = commInit(commSpec, FALSE)) == INVALID_HANDLE_VALUE)
	{
		fprintf(stderr, "Can't open comm channel using \"%s\".\n", commSpec);
		return(-1);
	}

	// make sure the programmer is in command mode
	abortCommand(cHand);

	// determine whether or not to provide visual feedback
	feedback = (_isatty(fileno(stderr)) && !_isatty(fileno(fp)));

	state = HEX_READ_PROMPT;
	while (state != HEX_READ_DONE) 
	{
		COMSTAT cs;
		BYTE c;
		DWORD cnt;

		ClearCommError(cHand, &cnt, &cs);
		if (state == HEX_READ_PROMPT)
		{
			// prompt the reader to send the first record
			c = 'R';
			if (!WriteFile(cHand, &c, 1, &cnt, NULL))
			{
				fprintf(stderr, "Error writing to comm port.\n");
				break;
			}
			state = HEX_READ_START;
		}
		else if (cs.cbInQue == 0)
			Yield();
		else if (ReadFile(cHand, &c, 1, &cnt, NULL) && (cnt > 0))
		{
			BYTE byteCnt;
			BYTE byteVal;
			BYTE charCnt;
			BYTE cksum;
			BYTE recType;
			WORD addr;

			if (state == HEX_READ_START)
			{
				if (c == ':')
				{
					fprintf(fp, "%c", c);
					state = HEX_READ_BYTECNT;
					cksum = 0;
					byteCnt = 0;
					charCnt = 2;
				}
				continue;
			}
			else if ((c == '\r') || (c == '\n'))
				continue;
			else if (!isxdigit(c))
			{
				fprintf(stderr, "Invalid character received, expecting hex.\n");
				c = NAK_CHAR;
				WriteFile(cHand, &c, 1, &cnt, NULL);
				break;
			}

			// echo the character
			fputc((WORD)c & 0xff, fp);

			switch(state)
			{
			case HEX_READ_BYTECNT:		// reading byte count
				byteCnt = (byteCnt * 16) + hexVal(c);
				if (--charCnt <= 0)
				{
					cksum += byteCnt;
					state = HEX_READ_ADDRESS;
					addr = 0;
					charCnt = 4;
				}
				break;

			case HEX_READ_ADDRESS:	// reading address
				addr = (addr * 16) + hexVal(c);
				if (--charCnt <= 0)
				{
					cksum += (((addr >> 8) & 0xff) + (addr & 0xff));
					state = HEX_READ_RECTYPE;
					recType = 0;
					charCnt = 2;
				}
				break;

			case HEX_READ_RECTYPE:	// reading record type
				recType = (recType * 16) + hexVal(c);
				if (--charCnt <= 0)
				{
					cksum += recType;
					byteVal = 0;
					if (byteCnt > 0)
					{
						state = HEX_READ_RECORD;
						charCnt = byteCnt * 2;
					}
					else
					{
						state = HEX_READ_CHKSUM;
						charCnt = 2;
					}
				}
				break;

			case HEX_READ_RECORD:	// reading record bytes
				byteVal = (byteVal * 16) + hexVal(c);
				if ((--charCnt & 0x01) == 0)
				{
					cksum += byteVal;
					byteVal = 0;
				}
				if (charCnt <= 0)
				{
					state = HEX_READ_CHKSUM;
					byteVal = 0;
					charCnt = 2;
				}
				break;

			case HEX_READ_CHKSUM:	// reading checksum
				byteVal = (byteVal * 16) + hexVal(c);
				if (--charCnt <= 0)
				{
					recCnt++;
					if (((cksum + byteVal) & 0xff) == 0)
					{
						fprintf(fp, "\n");
						c = ACK_CHAR;
						state = HEX_READ_START;
					}
					else
					{
						fprintf(stderr, "\nInvalid checksum on record %d\n  is %02x s/b %02x\n", recCnt, (~cksum + 1) & 0xff, byteVal);
						c = NAK_CHAR;
						state = HEX_READ_DONE;
					}
					if (recType == 0x01)
					{
						state = HEX_READ_DONE;
						c = NO_ACK;
					}
					else if ((state != HEX_READ_DONE) && feedback && ((recCnt % 16) == 0))
					{
						fputc('.', stderr);
						fflush(stderr);
					}
					if (c != NO_ACK)
						WriteFile(cHand, &c, 1, &cnt, NULL);
				}
				break;
			}
		}
	}
	commDeInit(cHand);
	return(state == HEX_READ_DONE);
}

/*
 ** validateHexRecord
 *
 * Scan an Intel hex record and validate its format.  Return
 * the record type if it is valid and return the length (in characters)
 * indirectly.  Otherwise, return an invalid record type.
 *
 */
static BYTE
validateHexRecord(BYTE *buf, int *lenp)
{
	int state;
	BYTE byteCnt = 0;
	BYTE byteVal = 0;
	BYTE charCnt = 0;
	BYTE cksum = 0;
	BYTE recType = 0;
	WORD addr = 0;

	state = HEX_READ_START;
	while (state != HEX_READ_DONE)
	{
		BYTE c;

		c = *buf++;
		if (state == HEX_READ_START)
		{
			if (c != ':')
				return(REC_TYPE_INVALID);

			state = HEX_READ_BYTECNT;
			cksum = 0;
			byteCnt = 0;
			charCnt = 2;
			continue;
		}

		// otherwise we need a hex digit
		if (!isxdigit(c))
			return(REC_TYPE_INVALID);
		c = hexVal(c);
		switch(state)
		{
		case HEX_READ_BYTECNT:		// reading byte count
			byteCnt = (byteCnt << 4) + c;
			if (--charCnt <= 0)
			{
				cksum += byteCnt;
				state = HEX_READ_ADDRESS;
				addr = 0;
				charCnt = 4;
			}
			break;

		case HEX_READ_ADDRESS:	// reading address
			addr = (addr << 4) + c;
			if (--charCnt <= 0)
			{
				cksum += (((addr >> 8) & 0xff) + (addr & 0xff));
				state = HEX_READ_RECTYPE;
				recType = 0;
				charCnt = 2;
			}
			break;

		case HEX_READ_RECTYPE:	// reading record type
			recType = (recType << 4) + c;
			if (--charCnt <= 0)
			{
				cksum += recType;
				byteVal = 0;
				if (byteCnt > 0)
				{
					state = HEX_READ_RECORD;
					charCnt = byteCnt * 2;
				}
				else
				{
					state = HEX_READ_CHKSUM;
					charCnt = 2;
				}
			}
			break;

		case HEX_READ_RECORD:	// reading record bytes
			byteVal = (byteVal << 4) + c;
			if ((--charCnt & 0x01) == 0)
			{
				cksum += byteVal;
				byteVal = 0;
			}
			if (charCnt <= 0)
			{
				state = HEX_READ_CHKSUM;
				byteVal = 0;
				charCnt = 2;
			}
			break;

		case HEX_READ_CHKSUM:	// reading checksum
			byteVal = (byteVal << 4) + c;
			if (--charCnt <= 0)
			{
				if (((cksum + byteVal) & 0xff) != 0)
					return(REC_TYPE_INVALID);
				state = HEX_READ_DONE;
			}
			break;

		default:	// this shouldn't ever happen
			return(REC_TYPE_INVALID);
		}
	}
	if (lenp != NULL)
		*lenp = (byteCnt * 2) + 11;
	return(recType);
}

/*
 ** writeFlash
 *
 * Read Intel hex records from the given file pointer and send them to
 * the device for writing to flash memory.
 *
 */
static int
writeFlash(FILE *fp, int verifyOnly)
{
	int stat = -1;
	HANDLE cHand;
	int state;
	int feedback;
	WORD recCnt = 0;
	char buf[80];
	long line = 0;

	if (fp == NULL)
		fp = stdin;

	if ((cHand = commInit(commSpec, FALSE)) == INVALID_HANDLE_VALUE)
	{
		fprintf(stderr, "Can't open comm channel using \"%s\".\n", commSpec);
		return(-1);
	}

	// make sure the programmer is in command mode
	abortCommand(cHand);

	// determine whether or not to provide visual feedback
	feedback = _isatty(fileno(stderr));
	state = HEX_WRITE_PROMPT;
	while (fgets(buf, sizeof(buf), fp) != NULL) 
	{
		COMSTAT cs;
		BYTE c;
		int len;
		DWORD cnt;
		BYTE recType;


		line++;

		// make sure the record is OK to send
		recType = validateHexRecord(buf, &len);
		if ((recType != REC_TYPE_DATA) && (recType != REC_TYPE_END))
		{
			fprintf(stderr, "Invalid hex record in file at line %lu -\n%s", line, buf);
			break;
		}
		if (len > MAX_REC_LEN)
		{
			fprintf(stderr, "\nOversized record in file at line %lu, len=%u.\n", line, len);
			break;
		}
		if (len <= 0)
			continue;
		buf[len] = '\0';


		if (state == HEX_WRITE_PROMPT)
		{
			// command the device to enter receive mode
			c = (verifyOnly ? 'V' : 'W');
			if (!WriteFile(cHand, &c, 1, &cnt, NULL))
			{
				fprintf(stderr, "\nError writing to comm port.\n");
				break;
			}
			state = HEX_WRITE_RECORDS;
		}

		// wait for previous transmission to complete
		while (1)
		{
			ClearCommError(cHand, &cnt, &cs);
			if (cs.cbOutQue == 0)
				break;
			Yield();
		}

		// send the record and an EOL
		if (!WriteFile(cHand, buf, (DWORD)len, &cnt, NULL) ||
				!WriteFile(cHand, "\r\n", 2, &cnt, NULL))
		{
			fprintf(stderr, "\nError writing hex record to comm port.\n");
			break;
		}

		if (recType == REC_TYPE_END)
			break;
		recCnt++;

		// wait for the ack before proceeding
		while (1)
		{
			char c;

			ClearCommError(cHand, &cnt, &cs);
			if ((cs.cbInQue != 0) && ReadFile(cHand, &c, 1, &cnt, NULL) && (cnt > 0))
			{
				if (c == ACK_CHAR)
					break;
				if (verifyOnly && (c == MISMATCH_CHAR))
				{
					fprintf(stderr, "\nDevice reports data mismatch, record %u.\n", recCnt);
					goto done;
				}
				if (c == NAK_CHAR)
				{
					fprintf(stderr, "\nDevice aborted transmission.\n");
					goto done;
				}
			}
			Yield();
		}

		if (feedback && ((recCnt % 16) == 0))
		{
			fputc('.', stderr);
			fflush(stderr);
		}
	}
	if (verifyOnly)
		fprintf(stderr, "\nFlash memory verified.\n");
	else
		fprintf(stderr, "\nFlash memory written.\n");
	stat = 0;
done:
	commDeInit(cHand);
	return(state == HEX_WRITE_DONE);
}
